# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from glance_store import backend
from glance_store import exceptions as store_exc
from oslo_config import cfg

from glare.common import exception
from glare.common import utils

CONF = cfg.CONF

error_map = [{'catch': store_exc.NotFound,
              'raise': exception.NotFound},
             {'catch': store_exc.UnknownScheme,
              'raise': exception.BadRequest},
             {'catch': store_exc.BadStoreUri,
              'raise': exception.BadRequest},
             {'catch': store_exc.Duplicate,
              'raise': exception.Conflict},
             {'catch': store_exc.StorageFull,
              'raise': exception.Forbidden},
             {'catch': store_exc.StorageWriteDenied,
              'raise': exception.Forbidden},
             {'catch': store_exc.Forbidden,
              'raise': exception.Forbidden},
             {'catch': store_exc.Invalid,
              'raise': exception.BadRequest},
             {'catch': store_exc.BadStoreConfiguration,
              'raise': exception.GlareException},
             {'catch': store_exc.RemoteServiceUnavailable,
              'raise': exception.BadRequest},
             {'catch': store_exc.HasSnapshot,
              'raise': exception.Conflict},
             {'catch': store_exc.InUseByStore,
              'raise': exception.Conflict},
             {'catch': store_exc.BackendException,
              'raise': exception.GlareException},
             {'catch': store_exc.GlanceStoreException,
              'raise': exception.GlareException}]


@utils.error_handler(error_map)
def save_blob_to_store(blob_id, blob, context, max_size,
                       store_type=None, verifier=None):
    """Save file to specified store type and return location info to the user

    :param store_type: type of the store, None means save to default store.
    :param blob_id: id of artifact
    :param blob: blob file iterator
    :param context: user context
    :param verifier:signature verified
    :return: tuple of values: (location_uri, size, checksums, metadata)
    """
    data = utils.LimitingReader(utils.CooperativeReader(blob), max_size)
    (location, size, md5checksum, metadata) = backend.add_to_backend(
        CONF, blob_id, data, 0, store_type, context, verifier)
    checksums = {"md5": md5checksum,
                 "sha1": data.sha1.hexdigest(),
                 "sha256": data.sha256.hexdigest()}
    return location, size, checksums


@utils.error_handler(error_map)
def load_from_store(uri, context):
    """Load file from store backend.

    :param uri: blob uri
    :param context: user context
    :return: file iterator
    """
    return backend.get_from_backend(uri=uri, context=context)[0]


@utils.error_handler(error_map)
def delete_blob(uri, context):
    """Delete blob from backend store

    :param uri: blob uri
    :param context: user context
    """
    return backend.delete_from_backend(uri, context)


@utils.error_handler(error_map)
def get_blob_size(uri, context):
    return backend.get_size_from_backend(uri, context)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_serialization import jsonutils
from six.moves import http_client
import webob.dec

from glare.api.v1 import api_version_request
from glare.i18n import _


versions_opts = [
    cfg.StrOpt('public_endpoint',
               help=_("""
Public url endpoint to use for Glare versions response.

This is the public url endpoint that will appear in the Glare
"versions" response. If no value is specified, the endpoint that is
displayed in the version's response is that of the host running the
API service. Change the endpoint to represent the proxy URL if the
API service is running behind a proxy. If the service is running
behind a load balancer, add the load balancer's URL for this value.

Services which consume this:
    * glare

Possible values:
    * None
    * Proxy URL
    * Load balancer URL

Related options:
    * None

""")),
]


CONF = cfg.CONF
CONF.register_opts(versions_opts)


class Controller(object):

    """A controller that reports which API versions are supported."""

    @staticmethod
    def index(req, is_multi):
        """Respond to a request for all OpenStack API versions.
        :param is_multi: defines if multiple choices should be response status
        or not
        :param req: user request object
        :return list of supported API versions
        """
        def build_version_object(max_version, min_version, status, path=None):
            url = CONF.public_endpoint or req.host_url
            return {
                'id': 'v%s' % max_version,
                'links': [
                    {
                        'rel': 'self',
                        'href': '%s/%s/' % (url, path) if path else
                        '%s/' % url,
                    },
                ],
                'status': status,
                'min_version': min_version,
                'version': max_version
            }

        microv_max = api_version_request.APIVersionRequest.max_version()
        microv_min = api_version_request.APIVersionRequest.min_version()
        version_objs = [build_version_object(microv_max.get_string(),
                                             microv_min.get_string(),
                                             'EXPERIMENTAL')]
        return_status = (http_client.MULTIPLE_CHOICES if is_multi else
                         http_client.OK)
        response = webob.Response(request=req,
                                  status=return_status,
                                  content_type='application/json')
        response.body = jsonutils.dump_as_bytes(dict(versions=version_objs))
        return response



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import jsonpatch
from oslo_log import log as logging

from glare.common import exception
from glare.common import policy
from glare.common import utils
from glare.db import artifact_api
from glare.i18n import _
from glare import locking
from glare.notification import Notifier
from glare.objects.meta import registry as glare_registry

LOG = logging.getLogger(__name__)


class Engine(object):
    """Engine is responsible for executing different helper operations when
    processing incoming requests from Glare API.
    Engine receives incoming data and does the following:
    - check basic policy permissions
    - requests artifact definition from registry
    - check access permission(ro, rw)
    - lock artifact for update if needed
    - pass data to base artifact to execute all business logic operations
    - notify other users about finished operation.
    Engine should not include any business logic and validation related
    to Artifacts. Engine should not know any internal details of Artifacts
    because it controls access to Artifacts in common.
    """

    registry = glare_registry.ArtifactRegistry
    registry.register_all_artifacts()
    lock_engine = locking.LockEngine(artifact_api.ArtifactLockApi())

    @classmethod
    def _get_schemas(cls, reg):
        if getattr(cls, 'schemas', None):
            pass
        else:
            schemas = {}
            for name, type_list in reg.obj_classes().items():
                type_name = type_list[0].get_type_name()
                schemas[type_name] = \
                    reg.get_artifact_type(type_name).gen_schemas()
            setattr(cls, 'schemas', schemas)
        return copy.deepcopy(cls.schemas)

    @classmethod
    def _get_artifact(cls, context, type_name, artifact_id,
                      read_only=False):
        """Return artifact for users

        Return artifact for reading/modification by users. Check
        access permissions and policies for artifact.
        """

        def _check_read_write_access(ctx, af):
            """Check if artifact can be modified by user

            :param ctx: user context
            :param af: artifact definition
            :raise Forbidden if access is not allowed
            """
            if not ctx.is_admin and ctx.tenant != af.owner or ctx.read_only:
                raise exception.Forbidden()

        def _check_read_only_access(ctx, af):
            """Check if user has read only access to artifact

            :param ctx: user context
            :param af: artifact definition
            :raise Forbidden if access is not allowed
            """
            private = af.visibility != 'public'
            if (private and
                    not ctx.is_admin and ctx.tenant != af.owner):
                # TODO(kairat): check artifact sharing here
                raise exception.Forbidden()

        artifact_type = Engine.registry.get_artifact_type(type_name)
        # only artifact is available for class users
        artifact = artifact_type.get(context, artifact_id)
        if read_only:
            _check_read_only_access(context, artifact)
            LOG.debug("Artifact %s acquired for read-only access", artifact_id)
        else:
            _check_read_write_access(context, artifact)
            LOG.debug("Artifact %s acquired for read-write access",
                      artifact_id)
        return artifact

    @classmethod
    def list_type_schemas(cls, context):
        policy.authorize("artifact:type_list", {}, context)
        return cls._get_schemas(cls.registry)

    @classmethod
    def show_type_schema(cls, context, type_name):
        policy.authorize("artifact:type_get", {}, context)
        schemas = cls._get_schemas(cls.registry)
        if type_name not in schemas:
            msg = _("Artifact type %s does not exist") % type_name
            raise exception.NotFound(message=msg)
        return schemas[type_name]

    @classmethod
    def create(cls, context, type_name, field_values):
        """Create new artifact in Glare"""
        action_name = "artifact:create"
        policy.authorize(action_name, field_values, context)
        artifact_type = cls.registry.get_artifact_type(type_name)
        # acquire version lock and execute artifact create
        af = artifact_type.create(context, field_values)
        # notify about new artifact
        Notifier.notify(context, action_name, af)
        # return artifact to the user
        return af.to_dict()

    @classmethod
    @lock_engine.locked(['type_name', 'artifact_id'])
    def update(cls, context, type_name, artifact_id, patch):
        """Update artifact with json patch.

        Apply patch to artifact and validate artifact before updating it
        in database. If there is request for visibility change or custom
        location change then call specific method for that.

        :param context: user context
        :param type_name: name of artifact type
        :param artifact_id: id of the artifact to be updated
        :param patch: json patch
        :return: updated artifact
        """

        def get_updates(af_dict, patch_with_upd):
            """Get updated values for artifact and json patch

            :param af_dict: current artifact definition as dict
            :param patch_with_upd: json-patch
            :return: dict of updated attributes and their values
            """

            try:
                af_dict_patched = patch_with_upd.apply(af_dict)
                diff = utils.DictDiffer(af_dict_patched, af_dict)

                # we mustn't add or remove attributes from artifact
                if diff.added() or diff.removed():
                    msg = _(
                        "Forbidden to add or remove attributes from artifact. "
                        "Added attributes %(added)s. "
                        "Removed attributes %(removed)s") % {
                        'added': diff.added(), 'removed': diff.removed()
                    }
                    raise exception.BadRequest(message=msg)

                return {key: af_dict_patched[key] for key in diff.changed()}

            except (jsonpatch.JsonPatchException,
                    jsonpatch.JsonPointerException,
                    KeyError) as e:
                raise exception.BadRequest(message=str(e))
            except TypeError as e:
                msg = _("Incorrect type of the element. Reason: %s") % str(e)
                raise exception.BadRequest(msg)

        artifact = cls._get_artifact(context, type_name, artifact_id)
        af_dict = artifact.to_dict()
        updates = get_updates(af_dict, patch)
        LOG.debug("Update diff successfully calculated for artifact %(af)s "
                  "%(diff)s", {'af': artifact_id, 'diff': updates})

        if not updates:
            return af_dict
        else:
            action = artifact.get_action_for_updates(context, artifact,
                                                     updates, cls.registry)
            action_name = "artifact:%s" % action.__name__
            policy.authorize(action_name, af_dict, context)
            modified_af = action(context, artifact, updates)
            Notifier.notify(context, action_name, modified_af)
            return modified_af.to_dict()

    @classmethod
    def get(cls, context, type_name, artifact_id):
        """Return artifact representation from artifact repo."""
        policy.authorize("artifact:get", {}, context)
        af = cls._get_artifact(context, type_name, artifact_id,
                               read_only=True)
        return af.to_dict()

    @classmethod
    def list(cls, context, type_name, filters, marker=None, limit=None,
             sort=None, latest=False):
        """Return list of artifacts requested by user.

        :param context: user context
        :param type_name: Artifact type name
        :param filters: filters that need to be applied to artifact
        :param marker: the artifact that considered as begin of the list
        so all artifacts before marker (including marker itself) will not be
        added to artifact list
        :param limit: maximum number of items in list
        :param sort: sorting options
        :param latest: flag that indicates, that only artifacts with highest
        versions should be returned in output
        :return: list of artifacts
        """
        policy.authorize("artifact:list", {}, context)
        artifact_type = cls.registry.get_artifact_type(type_name)
        # return list to the user
        af_list = [af.to_dict()
                   for af in artifact_type.list(context, filters, marker,
                                                limit, sort, latest)]
        return af_list

    @classmethod
    def delete(cls, context, type_name, artifact_id):
        """Delete artifact from glare"""
        af = cls._get_artifact(context, type_name, artifact_id)
        policy.authorize("artifact:delete", af.to_dict(), context)
        af.delete(context, af)
        Notifier.notify(context, "artifact.delete", af)

    @classmethod
    @lock_engine.locked(['type_name', 'artifact_id'])
    def add_blob_location(cls, context, type_name,
                          artifact_id, field_name, location, blob_meta):
        af = cls._get_artifact(context, type_name, artifact_id)
        action_name = 'artifact:set_location'
        policy.authorize(action_name, af.to_dict(), context)
        modified_af = af.add_blob_location(context, af, field_name, location,
                                           blob_meta)
        Notifier.notify(context, action_name, modified_af)
        return modified_af.to_dict()

    @classmethod
    @lock_engine.locked(['type_name', 'artifact_id'])
    def add_blob_dict_location(cls, context, type_name, artifact_id,
                               field_name, blob_key, location, blob_meta):
        af = cls._get_artifact(context, type_name, artifact_id)
        action_name = 'artifact:set_location'
        policy.authorize(action_name, af.to_dict(), context)
        modified_af = af.add_blob_dict_location(context, af, field_name,
                                                blob_key, location, blob_meta)
        Notifier.notify(context, action_name, modified_af)
        return modified_af.to_dict()

    @classmethod
    @lock_engine.locked(['type_name', 'artifact_id'])
    def upload_blob(cls, context, type_name, artifact_id, field_name, fd,
                    content_type):
        """Upload Artifact blob"""
        af = cls._get_artifact(context, type_name, artifact_id)
        action_name = "artifact:upload"
        policy.authorize(action_name, af.to_dict(), context)
        modified_af = af.upload_blob(context, af, field_name, fd, content_type)
        Notifier.notify(context, action_name, modified_af)
        return modified_af.to_dict()

    @classmethod
    @lock_engine.locked(['type_name', 'artifact_id'])
    def upload_blob_dict(cls, context, type_name, artifact_id, field_name,
                         blob_key, fd, content_type):
        """Upload Artifact blob to dict"""
        af = cls._get_artifact(context, type_name, artifact_id)
        action_name = "artifact:upload"
        policy.authorize(action_name, af.to_dict(), context)
        modified_af = af.upload_blob_dict(context, af, field_name, blob_key,
                                          fd, content_type)
        Notifier.notify(context, action_name, modified_af)
        return modified_af.to_dict()

    @classmethod
    def download_blob(cls, context, type_name, artifact_id, field_name):
        """Download blob from artifact"""
        af = cls._get_artifact(context, type_name, artifact_id,
                               read_only=True)
        policy.authorize("artifact:download", af.to_dict(), context)
        return af.download_blob(context, af, field_name)

    @classmethod
    def download_blob_dict(cls, context, type_name, artifact_id, field_name,
                           blob_key):
        """Download blob from artifact"""
        af = cls._get_artifact(context, type_name, artifact_id,
                               read_only=True)
        policy.authorize("artifact:download", af.to_dict(), context)
        return af.download_blob_dict(context, af, field_name, blob_key)



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import six

from oslo_log import log as logging

from glare.i18n import _LI

LOG = logging.getLogger(__name__)


class LockApiBase(object):
    """Lock Api Base class that responsible for acquiring/releasing locks
    """

    def create_lock(self, context, lock_key):
        """Acquire lock for current user

        :param context user context
        :param lock_key: unique lock identifier that defines lock scope
        :return lock internal identifier
        """
        raise NotImplementedError()

    def delete_lock(self, context, lock_id):
        """Delete acquired user lock

        :param context: user context
        :param lock_id: lock internal identifier
        :return:
        """
        raise NotImplementedError()


class Lock(object):
    """Object that stores lock context for users. This class is internal
    and used only for Lock Engine. So users shouldn't use this class directly
    """

    def __init__(self, context, lock_id, lock_key, release_method):
        """Initialize lock context"""
        self.context = context
        self.lock_id = lock_id
        self.lock_key = lock_key
        self.release = release_method

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # TODO(kairat) catch all exceptions here
        self.release(self)


class LockEngine(object):
    """Glare lock engine.
    Defines how artifact updates must be synchronized with each other. When
    some user obtains lock  for the same piece of data then other user cannot
    request that lock and get Conflict error.
    This little engine also allows to encapsulate lock logic in one place so
    we can potentially add tooz functionality in future to Glare. Right now
    there are troubles with locks in Galera (especially in mysql) and zookeeper
    requires additional work from IT engineers. So we need support production
    ready DB locks in our implementation.
    """

    MAX_LOCK_LENGTH = 255

    def __init__(self, lock_api):
        """Initialize lock engine with some lock api

        :param lock_api: api that allows to create/delete locks. It must be
        db_api but it might be replaced with DLM in near future.
        """
        self.lock_api = lock_api

    def acquire(self, context, lock_key):
        """Acquire lock to update whole artifact

        Acquire lock to update artifact. If there is some other
        lock for the same artifact then raise Conflict Error.

        :param context: user context
        :param lock_key: lock key
        :return: lock definition
        """
        if lock_key is not None and len(lock_key) < self.MAX_LOCK_LENGTH:
            lock_id = self.lock_api.create_lock(context, lock_key)
            LOG.info(_LI("Lock %(lock_id)s acquired for lock_key "
                         "%(lock_key)s"),
                     {'lock_id': lock_id, 'lock_key': lock_key})
        else:
            lock_id = None
            LOG.info(_LI("No lock for lock_key %s"), lock_key)

        return Lock(context, lock_id, lock_key, self.release)

    def release(self, lock):
        if lock.lock_id is not None:
            self.lock_api.delete_lock(lock.context, lock.lock_id)
            LOG.info(_LI("Lock %(lock_id)s released for lock_key %(key)s"),
                     {'lock_id': lock.lock_id, 'key': lock.lock_key})

    def locked(self, lock_name_parameters):
        """Synchronization decorator.
        :param list lock_name_parameters: List of parameters that will be used
        as part of lock name
        :returns: function that locks artifact by specified parameters
        """
        def wrap(f):
            @six.wraps(f)
            def wrapped(*a, **k):
                call_args = inspect.getcallargs(f, *a, **k)
                lock_key = ''
                for param in lock_name_parameters:
                    lock_key += str(call_args[param]) + ':'
                context = call_args.get('context')
                with self.acquire(context, lock_key):
                    res = f(*a, **k)
                return res
            return wrapped
        return wrap



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Common database interface for all objects"""


class BaseDBAPI(object):

    def __init__(self, cls):
        self.type = cls.get_type_name()
        self.cls = cls

    def create(self, context, values):
        """Create new artifact in db and return dict of values to the user

        :param context: user context
        :param values: dict of values that needs to be saved to db
        :return: dict of created values
        """
        raise NotImplementedError()

    def update(self, context, artifact_id, values):
        """Update artifact values in database

        :param artifact_id: id of artifact that needs to be updated
        :param context: user context
        :param values: values that needs to be updated
        :return: dict of updated artifact values
        """
        raise NotImplementedError()

    def get(self, context, artifact_id):
        """Return artifact values from database

        :param context: user context
        :param artifact_id: id of the artifact
        :return: dict of artifact values
        """
        raise NotImplementedError()

    def delete(self, context, artifact_id):
        """Delete artifacts from db

        :param context: user context
        :param artifact_id: id of artifact that needs to be deleted
        :return: dict for deleted artifact value
        """
        raise NotImplementedError()

    def list(self, context, filters, marker, limit, sort, latest):
        """List artifacts from db

        :param context: user request context
        :param filters: filter conditions from url
        :param marker: id of first artifact where we need to start
        artifact lookup
        :param limit: max number of items in list
        :param sort: sort conditions
        :param latest: flag that indicates, that only artifacts with highest
        versions should be returned in output
        :return: list of artifacts. Each artifact is represented as dict of
        values.
        """
        raise NotImplementedError()



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_versionedobjects import fields

from glare.i18n import _
from glare.objects.meta import fields as glare_fields

LOG = logging.getLogger(__name__)


class Validator(object):
    """Common interface for all validators"""

    def validate(self, value):
        raise NotImplementedError()

    def get_allowed_types(self):
        raise NotImplementedError()

    def check_type_allowed(self, field_type):
        if not issubclass(field_type, self.get_allowed_types()):
            # try to check if field_type is correct
            # in case of element_type passed
            allowed_field_types = tuple(type(field.AUTO_TYPE)
                                        for field in self.get_allowed_types()
                                        if hasattr(field, 'AUTO_TYPE'))
            if not issubclass(field_type, allowed_field_types):
                raise TypeError(
                    _("%(type)s is not allowed for validator "
                      "%(val)s. Allowed types are %(allowed)s.") % {
                        "type": str(field_type),
                        "val": str(self.__class__),
                        "allowed": str(self.get_allowed_types())})

    def to_jsonschema(self):
        return {}

    def __call__(self, value):
        try:
            self.validate(value)
        except ValueError:
            raise
        except TypeError as e:
            # we are raising all expected ex Type Errors as ValueErrors
            LOG.exception(e)
            raise ValueError(encodeutils.exception_to_unicode(e))


class UUID(Validator):
    def get_allowed_types(self):
        return fields.StringField,

    def validate(self, value):
        pass

    def to_jsonschema(self):
        return {'pattern': ('^([0-9a-fA-F]){8}-([0-9a-fA-F]){4}-([0-9a-fA-F])'
                            '{4}-([0-9a-fA-F]){4}-([0-9a-fA-F]){12}$')}


class AllowedValues(Validator):
    def __init__(self, allowed_values):
        self.allowed_values = allowed_values

    def get_allowed_types(self):
        return fields.StringField,

    def validate(self, value):
        if value not in self.allowed_values:
            raise ValueError(_("Value must be one of the following: %s") %
                             ', '.join(self.allowed_values))

    def to_jsonschema(self):
        return {'enum': self.allowed_values + [None]}


class Version(Validator):
    def get_allowed_types(self):
        return glare_fields.VersionField

    def validate(self, value):
        pass

    def to_jsonschema(self):
        return {'pattern': ('/^([0-9]+)\.([0-9]+)\.([0-9]+)(?:-([0-9A-Za-z-]'
                            '+(?:\.[0-9A-Za-z-]+)*))?(?:\+[0-9A-Za-z-]+)?$/')}


class SizeValidator(Validator):
    def __init__(self, size):
        self.size = size


class MaxStrLen(SizeValidator):
    def get_allowed_types(self):
        return fields.StringField,

    def validate(self, value):
        l = len(value)
        if l > self.size:
            raise ValueError(
                _("String length must be less than  %(size)s. "
                  "Current size: %(cur)s") % {'size': self.size,
                                              'cur': l})

    def to_jsonschema(self):
        return {'maxLength': self.size}


class MinStrLen(SizeValidator):
    def get_allowed_types(self):
        return fields.StringField,

    def validate(self, value):
        l = len(value)
        if l < self.size:
            raise ValueError(
                _("String length must be more than  %(size)s. "
                  "Current size: %(cur)s") % {'size': self.size,
                                              'cur': l})

    def to_jsonschema(self):
        return {'minLength': self.size}


class ForbiddenChars(Validator):
    def __init__(self, forbidden_chars):
        self.forbidden_chars = forbidden_chars

    def get_allowed_types(self):
        return fields.StringField,

    def validate(self, value):
        for fc in self.forbidden_chars:
            if fc in value:
                raise ValueError(
                    _("Forbidden character %(char) found in string "
                      "%(string)s")
                    % {"char": fc, "string": value})


class MaxSize(SizeValidator):

    def validate(self, value):
        l = len(value)
        if l > self.size:
            raise ValueError(
                _("Number of items must be less than  "
                  "%(size)s. Current size: %(cur)s") %
                {'size': self.size, 'cur': l})

    def to_jsonschema(self):
        return {'maxItems': self.size}


class MaxDictSize(MaxSize):

    def get_allowed_types(self):
        return glare_fields.Dict

    def to_jsonschema(self):
        return {'maxProperties': self.size}


class MaxListSize(MaxSize):

    def get_allowed_types(self):
        return glare_fields.List

    def to_jsonschema(self):
        return {'maxItems': self.size}


class MaxNumberSize(SizeValidator):
    def validate(self, value):
        if value > self.size:
            raise ValueError("Number is too big: %s. Max allowed number is "
                             "%s" % (value, self.size))

    def get_allowed_types(self):
        return fields.IntegerField, fields.FloatField

    def to_jsonschema(self):
        return {'maximum': self.size}


class MinNumberSize(SizeValidator):
    def validate(self, value):
        if value < self.size:
            raise ValueError("Number is too small: %s. Min allowed number is "
                             "%s" % (value, self.size))

    def get_allowed_types(self):
        return fields.IntegerField, fields.FloatField

    def to_jsonschema(self):
        return {'minumum': self.size}


class Unique(Validator):
    def get_allowed_types(self):
        return glare_fields.List,

    def validate(self, value):
        if len(value) != len(set(value)):
            raise ValueError(_("List items %s must be unique.") % value)

    def to_jsonschema(self):
        return {'unique': True}


class AllowedListValues(Validator):
    def __init__(self, allowed_values):
        self.allowed_items = allowed_values

    def get_allowed_types(self):
        return glare_fields.List,

    def validate(self, value):
        for item in value:
            if item not in self.allowed_items:
                raise ValueError(
                    _("Value %(item)s is not allowed in list. "
                      "Allowed list values: %(allowed)s") %
                    {"item": item,
                     "allowed": self.allowed_items})

    def to_jsonschema(self):
        return {'enum': self.allowed_items}


class AllowedDictKeys(Validator):
    def __init__(self, allowed_keys):
        self.allowed_items = allowed_keys

    def get_allowed_types(self):
        return glare_fields.Dict,

    def validate(self, value):
        for item in value:
            if item not in self.allowed_items:
                raise ValueError(_("Key %(item)s is not allowed in dict. "
                                   "Allowed key values: %(allowed)s") %
                                 {"item": item,
                                  "allowed": ', '.join(self.allowed_items)})

    def to_jsonschema(self):
        return {
            'properties': {prop: {} for prop in self.allowed_items},
        }


class RequiredDictKeys(Validator):
    def __init__(self, required_keys):
        self.required_items = required_keys

    def get_allowed_types(self):
        return glare_fields.Dict,

    def validate(self, value):
        for item in self.required_items:
            if item not in value:
                raise ValueError(_("Key \"%(item)s\" is required in property "
                                   "dictionary: %(value)s.") %
                                 {"item": item,
                                  "value": ''.join(
                                      '{}:{}, '.format(key, val)
                                      for key, val in six.iteritems(value))})

    def to_jsonschema(self):
        return {'required': list(self.required_items)}


class MaxDictKeyLen(SizeValidator):
    def get_allowed_types(self):
        return glare_fields.Dict,

    def validate(self, value):
        for key in value:
            if len(str(key)) > self.size:
                raise ValueError(_("Dict key length %(key)s must be less than "
                                   "%(size)s.") % {'key': key,
                                                   'size': self.size})


class ElementValidator(Validator):
    def __init__(self, validators):
        self.validators = validators


class ListElementValidator(ElementValidator):
    def get_allowed_types(self):
        return glare_fields.List,

    def validate(self, value):
        for v in value:
            for validator in self.validators:
                validator(v)


class DictElementValidator(ElementValidator):
    def get_allowed_types(self):
        return glare_fields.Dict,

    def validate(self, value):
        for v in six.itervalues(value):
            for validator in self.validators:
                validator(v)
